(*                                                              *)
(*              Copyright 1989 fred brooks LogicTek             *)
(*                                                              *)
(*   First Release                      12/8/87-FGB             *)
(*   Added drive config data routines, clean up VBL code,       *)
(*   Remote time option                 12/17/89-FGB            *)
(*                                                              *)

IMPLEMENTATION MODULE NETWORK ;

(* --------------------------------------------------------------------------

               NETWORK : TWO CPU NETWORK FOR TDI Modula-2/ST

   --------------------------------------------------------------------------*)

(*$T-,$S-,$A+ *)

FROM SYSTEM IMPORT ADDRESS, ADR, SETREG, CODE, REGISTER ,BYTE ,TSIZE;
FROM BIOS   IMPORT BPB ,BConStat ,BConIn, BCosStat, BConOut, Device,
                   MediaChange,MCState,GetBPB,RWAbs,RW,DriveSet,DriveMap;
FROM XBIOS  IMPORT SuperExec,IORec,IORECPTR,IOREC,SerialDevice,
                   GetDateTime,ScreenPhysicalBase;
FROM GEMDOS IMPORT TermRes,Open,Close,SetDate,SetTime ;
IMPORT             GEMDOS;
FROM ASCII  IMPORT SYN,STX,SOH,BEL,CR,LF,ESC;

CONST
  MaxSeq          = 1;
  Maxdrives       = 31; (* number of disk drives minus 1 *)
  recsize         = 511;
  retry           = 5;
  MAGIC           = 314159;
  Memdrive        = 31;
  chanwait        = 3;
  debug           = FALSE;
  trace           = FALSE;
  TITLE           = "MX2NET Version 2.0   Fred Brooks  UUCP crash!fgbrooks";

  (* Because we dont know what registers the BIOS is using we must use
     the following opcodes to save the registers *)
  MOVEMDEC = 48E7H ;    (* 68000 opcode for MOVEM <regs>,-(A7) *)
  MOVEMINC = 4CDFH ;    (* 68000 opcode for MOVEM (A7)+,<regs> *)
  SAVEREGS = 07FFCH ;   (* Registers D1..A5 for DEC *)
  RESTREGS = 03FFEH ;   (* Registers D1..A5 for INC *)
  RTS = 04E75H ;        (* 68000 return from subroutine opcode *)

TYPE
  (* Procedure types to mimic correct sequence for "C" BIOS routines *)

  CBPBProc     = PROCEDURE ( CARDINAL ) ;
  CMediaChProc = PROCEDURE ( CARDINAL ) ;
  CRWAbsProc   = PROCEDURE ( CARDINAL, CARDINAL, CARDINAL, ADDRESS, CARDINAL );
  MIDIbuffer   = ARRAY [0..1023] OF CARDINAL;
  SequenceNr   = [0..MaxSeq];
  message      = ARRAY [0..recsize] OF BYTE;
  message1     = ARRAY [0..17] OF BYTE;
  FrameKind    = (ack,data,callreq,callaccp,clearreq,clearconf,
                 resetreq,diag);
  DataKind     = (rdmediareq,rdmediaconf,rdbpbreq,rdbpbconf,
                 rdrwabsreq,rdrwabsconf,memreq,memconf,timereq,timeconf);
  evtype       = (framearrival,cksumerr,timeout,hostready,reset,nothing);
  channel      = (none,local,remote);

  frame        = RECORD
                 syn    :       CHAR; (* these are sync chars *)
                 stx    :       CHAR; (* for the frames       *)
                 kind   :       FrameKind;
                 seq    :       SequenceNr;
                 ack    :       SequenceNr;
                 cmd    :       DataKind;
                 rw     :       CARDINAL; (* read or write data *)
                 recno  :       CARDINAL; (* sector for data*)
                 d0     :       LONGCARD; (* data return variable *)
                 info   :       message;
                 cksum  :       CARDINAL;
               END;

  framecptr    = POINTER TO framecmd;

  framecmd     = RECORD
                 syn    :       CHAR; (* these are sync chars *)
                 stx    :       CHAR; (* for the frames       *)
                 kind   :       FrameKind;
                 seq    :       SequenceNr;
                 ack    :       SequenceNr;
                 cmd    :       DataKind;
                 rw     :       CARDINAL; (* read or write data *)
                 recno  :       CARDINAL; (* sector for data*)
                 d0     :       LONGCARD; (* data return variable *)
                 info   :       message1;
                 cksum  :       CARDINAL;
               END;

  control     = RECORD
                 magic          :       LONGCARD;
                 reset          :       BOOLEAN;
                 networkactive  :       BOOLEAN;
                 remotedrive    :       CARDINAL;
                 drivemap       :       DriveSet;
                nextframetosend :      SequenceNr;
                frameexpected   :      SequenceNr;
                sendreset       :      BOOLEAN;
               END;

  netmap      = RECORD
                 Remote         :       CARDINAL;
                 Local          :       CARDINAL;
                 Write          :       BOOLEAN;
                END;

  frameptr      =       POINTER TO ARRAY [0..1024] OF BYTE;

VAR


  (* BIOS variables : These can only be accessed with the 68000 in supervisor
     mode. The Modula-2 language allows you to fix the location of variables *)

  HDBPB     [0472H] : ADDRESS ;       (* hard disk get Bios Parameter Block *)
  HDRWAbs   [0476H] : ADDRESS ;       (* hard disk read/write abs   *)
  HDMediaCh [047EH] : ADDRESS ;       (* hard disk media change     *)
  EvtCritic [0404H] : ADDRESS ;       (* evt_critic *)
  DriveBits [04C2H] : SET OF [0..31]; (* disk drives present map    *)
  flock     [043EH] : LONGCARD;       (* disk access in progress    *)
  hz200     [04baH] : LONGCARD;       (* 200hz clock counter        *)

  NetBits           : SET OF [0..31];
  Dptr              : DriveSet;       (* save original drive map    *)
  Mptr              : LONGCARD;
  charcount,framesize,cksum,recframesize,sndframesize,
  SIZEframe,SIZEframecmd                                : CARDINAL;

  networkconnect          :   BOOLEAN; (* DCD = 1 TRUE  *)
  gotframe                :   BOOLEAN;
  framebufferfull         :   BOOLEAN;
  cleartosend             :   BOOLEAN;
  readytosend             :   BOOLEAN;
  requesttosend           :   BOOLEAN;
  framewaiting            :   BOOLEAN;
  OK,installed            :   BOOLEAN;
  gotmediach              :   ARRAY [0..Maxdrives] OF BOOLEAN;
  gotbpb                  :   ARRAY [0..Maxdrives] OF BOOLEAN;
  networkerror            :   BOOLEAN;
  shortframe              :   BOOLEAN;
  vblLock                 :   BOOLEAN;
  rwabsLock               :   BOOLEAN;
  TIMESET                 :   BOOLEAN;
  OneTime                 :   BOOLEAN;
  ChannelLock             :   channel;
  NetMap                  :   ARRAY [0..Maxdrives] OF netmap;
  NetInfo                 :   ARRAY [0..128] OF CHAR;
  statptr                 :   POINTER TO stat;

  sframe,rframe,SFRAME,RFRAME                   :   frame;
  rframeptr                                     :   frameptr;
  framecmdptr                                   :   framecptr;
  sframecmdptr                                  :   framecptr;
  event                                         :   evtype;
  C                                             :   control;
  S                                             :   stat;
  recchar                                       :   LONGCARD;
  result,i,i1,mediacount,handle                 :   INTEGER;
  D0ptr                                         :   POINTER TO LONGCARD;
  wsector,drvnr,d,R                             :   CARDINAL;
  rbuffer                                       :   MIDIbuffer;
  rbptr,kbdiorec                                :   IORECPTR;
  numBytes,sec,min,hour,time,count              :   LONGCARD ;
  status                                        :   LONGINT ;
  sframeptr                                     :   frameptr;

  (* The following are saved copies of the BIOS variables so that the real
     hard disk routines can be called if a hard disk access is requested. *)

  SaveHDBPB      : CBPBProc ;     (* hard disk get Bios Parameter Block *)
  SaveHDRWAbs    : CRWAbsProc ;   (* hard disk read/write abs *)
  SaveHDMediaCh  : CMediaChProc ; (* hard disk media change *)
  SaveCritic     : PROC;

  (* NETWORK control *)

  NetworkBPB  : ARRAY [0..Maxdrives] OF BPB ; (* BIOS Parameter block for NETWORK *)


PROCEDURE inc(VAR k: SequenceNr);   (* increment k circulary *)
BEGIN
        IF k<MaxSeq THEN INC(k) ELSE k:=0 END;
END     inc;

MODULE  NETBIOS;
IMPORT  getfromremote,frameptr,ADDRESS,CODE,MOVEMDEC,SAVEREGS,status,
        Memdrive,statptr,S,MOVEMINC,RESTREGS,NetBits,wsector,DataKind,
        C,NetMap,resetnewdisk,channel,ADR,rwabsLock,frame,
        networkerror,SETREG,SaveHDBPB,SaveHDRWAbs,SaveHDMediaCh,TSIZE,
        newdisk,gotbpb,gotmediach,MCState,NetworkBPB,BPB;

EXPORT  RDRWAbs,RDMediaCh,RDBPB;

VAR     i3                              :       CARDINAL;
        bpbptr,nbpbptr                  :       frameptr;
        nframe1                         :       frame;

PROCEDURE MoveMemory ( From, To : ADDRESS ; Bytes : LONGCARD ) ;
(* This routine shows how time critical portions of code can be optimised to
   run faster. It relys on the code generation rules of the compiler which 
   can be checked by dis-assembling the link file with DecLnk.*)

CONST
  MOVEB = 12D8H ;       (*      MOVE.B  (A0)+,(A1)+     *)
  MOVEL = 22D8H ;       (*      MOVE.L  (A0)+,(A1)+     *)
  A0    = 0+8 ;         (* register A0 *)
  A1    = 1+8 ;         (* register A1 *)

BEGIN
  SETREG(A0,From) ;             (* load From pointer into A0 *)
  SETREG(A1,To) ;               (* load To pointer into A1 *)
  
  IF ( ODD(From) OR ODD(To) ) THEN      (* must do bytes *)
    WHILE ( Bytes <> 0 ) DO
      CODE(MOVEB) ;
      DEC(Bytes) ;
    END ;
  ELSE (* even addresses so can do long moves *)
    WHILE ( Bytes > 3 ) DO
      CODE(MOVEL) ;
      DEC(Bytes,4) ;
    END ;
    WHILE ( Bytes <> 0 ) DO
      CODE(MOVEB) ;             (* clean up remainder *)
      DEC(Bytes) ;
    END ;
  END ;
END MoveMemory ;

(* The following procedures mimic the disk handling routines called by the
   BIOS. Their procedure declarations have been written to mimic the "C"
   calling sequence. *)

PROCEDURE RDRWAbs ( device, RecordNum, SectorCount : CARDINAL ;
                    Buffer : ADDRESS ; Flag : CARDINAL ) ;
(* NB. It is assumed that GEMDOS wont call this routine with out of range
   parameters *)
CONST D0 = 0 ;
BEGIN
  CODE(MOVEMDEC,SAVEREGS) ;     (* save registers on stack *)
  status := 0;
  IF (device=Memdrive) AND (RecordNum=0) THEN (* get network stats *)
    statptr:=Buffer;
    statptr^:=S;
    CODE(MOVEMINC,RESTREGS) ;     (* Restore registers from stack *)
    RETURN;
  END;
  IF device IN NetBits THEN (* is NETWORK channel *)
    IF ( Flag = 0 ) OR ( Flag = 2 ) (* read *)  THEN
      FOR wsector:=0 TO (SectorCount-1) DO
           C.remotedrive:=NetMap[device].Remote;
           nframe1.d0:=LONGCARD(NetMap[device].Remote);
           nframe1.recno:=RecordNum+wsector;
           nframe1.rw:=Flag; (* read *)
           resetnewdisk;
           IF getfromremote(rdrwabsreq,rdrwabsconf,nframe1,local) THEN
              MoveMemory(ADR(nframe1.info),Buffer+ADDRESS(wsector)*512,
                         512);
              status:=0;
           ELSE
              status:=(-11);
           END; (* if *)
       END; (* for *)
    IF networkerror THEN C.sendreset:=TRUE END; (* send network reset to remote cpu *)
      SETREG(D0,status) ;
    ELSIF ( Flag = 1 ) OR ( Flag = 3 ) THEN (* write *)
       IF NetMap[device].Write THEN
         FOR wsector:=0 TO (SectorCount-1) DO
           C.remotedrive:=NetMap[device].Remote;
           nframe1.d0:=LONGCARD(NetMap[device].Remote);
           nframe1.recno:=RecordNum+wsector;
           nframe1.rw:=Flag; (* write *)
           resetnewdisk;
           MoveMemory(Buffer+ADDRESS(wsector)*512,ADR(nframe1.info),512);
           IF getfromremote(rdrwabsreq,rdrwabsconf,nframe1,local) THEN
              status:=0;
           ELSE
              status:=(-10);
           END;
         END; (* for *)
    IF networkerror THEN C.sendreset:=TRUE END; (* send network reset to remote cpu *)
       ELSE
         status:=(-13); (* write protect *)
       END;
      SETREG(D0,status) ;
    ELSE
      SETREG(D0,LONGINT(-3)) ;
    END ;
  ELSE (* not NETWORK *)
    rwabsLock:=TRUE;
    SaveHDRWAbs (device,RecordNum,SectorCount,Buffer,Flag) ;
    rwabsLock:=FALSE;
  END ;
  CODE(MOVEMINC,RESTREGS) ;     (* Restore registers from stack *)
END RDRWAbs ;

PROCEDURE RDMediaCh ( device : CARDINAL ) ;
CONST D0 = 0 ;
BEGIN
  CODE(MOVEMDEC,SAVEREGS) ;     (* save registers on stack *)
  IF device IN NetBits THEN (* is NETWORK channel *)
    C.remotedrive:=NetMap[device].Remote;
    nframe1.d0:=LONGCARD(NetMap[device].Remote);
    IF newdisk() THEN
       gotmediach[NetMap[device].Remote]:=FALSE;
       gotbpb[NetMap[device].Remote]:=FALSE;
    END;
    IF (NOT gotmediach[NetMap[device].Remote]) THEN
     IF getfromremote(rdmediareq,rdmediaconf,nframe1,local) THEN
        gotmediach[NetMap[device].Remote]:=TRUE;
        IF nframe1.d0=1 THEN nframe1.d0:=2 END;
        SETREG(D0,nframe1.d0) ;    (* "C" uses D0 as return location *)
     ELSE
        SETREG(D0,Changed);
     END;
    ELSE
       SETREG(D0,NoChange) ;    (* "C" uses D0 as return location *)
    END; 
  ELSE (* not NETWORK *)
    rwabsLock:=TRUE;
    SaveHDMediaCh(device) ;
    rwabsLock:=FALSE;
  END;
  CODE(MOVEMINC,RESTREGS) ;     (* Restore registers from stack *)
END RDMediaCh ;

PROCEDURE RDBPB ( device : CARDINAL ) ;
CONST D0 = 0 ;
BEGIN
  CODE(MOVEMDEC,SAVEREGS) ;     (* save registers on stack *)
  IF device IN NetBits THEN (* is NETWORK channel *)
    C.remotedrive:=NetMap[device].Remote;
    nframe1.d0:=LONGCARD(NetMap[device].Remote);
    IF newdisk() THEN
      gotbpb[NetMap[device].Remote]:=FALSE;
      gotmediach[NetMap[device].Remote]:=FALSE;
    END;
    IF (NOT gotbpb[NetMap[device].Remote]) THEN
     IF getfromremote(rdbpbreq,rdbpbconf,nframe1,local) THEN
       gotbpb[NetMap[device].Remote]:=TRUE;
       bpbptr:=ADR(nframe1.info);
       nbpbptr:=ADR(NetworkBPB[NetMap[device].Remote]);
       FOR i3:=0 TO TSIZE(BPB)-1 DO
           nbpbptr^[i3]:=bpbptr^[i3];    
       END;
       resetnewdisk;
       SETREG(D0,ADR(NetworkBPB[NetMap[device].Remote])); (* D0 returns address of the BPB *)
     ELSE
       SETREG(D0,0);
     END;
    ELSE
       SETREG(D0,ADR(NetworkBPB[NetMap[device].Remote])); (* D0 returns address of the BPB *)
    END; 
    IF networker